import sqlite3
import numpy as np
import filelock
from os import listdir
import re
import os


class OMPSTomographyJobs:
    def __init__(self, database, table, db_folder):
        self.database = db_folder + database
        self.table = table
        self.database_lock = db_folder + database + '.lock'
        self.lock = filelock.FileLock(self.database_lock, timeout=60)
        if not os.path.exists(db_folder):
            os.makedirs(db_folder)

    def make_new_table(self):
        conn = sqlite3.connect(self.database)
        c = conn.cursor()

        with self.lock:
            c.execute("CREATE TABLE " + str(self.table) + " (profileID PRIMARY KEY, isReady, isProcessing, isFinished, failed, priority DEFAULT(10))")

            conn.commit()
            conn.close()

    def add_orbit_to_table(self, orbit, priority=10):
        try:
            # Convert orbit and priority to be in iterable format of the same length
            if hasattr(orbit, '__len__') & ~hasattr(priority, '__len__'):
                priority = np.ones(np.shape(orbit))*priority
            if ~hasattr(orbit, '__len__'):
                orbit = [orbit]
            if ~hasattr(priority, '__len__'):
                priority = [priority]

            self.__add_entry_to_table(orbit, priority)
        except:
            print('could not add orbit ' + str(orbit) + ' to the database')

    def __add_entry_to_table(self, orbit, priority=10):
        try:
            conn = sqlite3.connect(self.database)
            c = conn.cursor()
        except:
            print('Could not connect to the sqlite database')

        with self.lock:
            for o, p in zip(orbit, priority):
                job = (int(o), 1, 0, 0, 0, int(p))
                c.execute('INSERT INTO ' + self.table + ' VALUES (?,?,?,?,?,?)',job)

            conn.commit()
            conn.close()

    def add_entry_to_table_from_txt(self, filename):

        temp = np.ndfromtxt(filename, dtype=int)

        if np.shape(temp)[1] == 3:
            for idx, t in enumerate(temp):
                self.add_entry_to_table(t[0],t[1],t[2])
        elif np.shape(temp)[1] == 2:
            for t in temp:
                self.add_orbit_to_table(t[0], priority=t[1])
        elif np.shape(temp)[1] == 1:
            for orbit in temp:
                self.add_orbit_to_table(orbit)

    def get_next_entry(self):
        conn = sqlite3.connect(self.database)
        c = conn.cursor()
        data = c.execute('SELECT * FROM ' + self.table + ' WHERE isReady=1 AND isProcessing=0 AND isFinished=0 ORDER BY priority DESC LIMIT 1')
        temp = data.fetchall()
        conn.close()

        if len(temp) == 0:
            orbit = None
            image = None
        else:
            temp = temp[0]
            orbit = int(temp[0])

        return orbit

    def set_entry_as_processing(self, orbit):
        conn = sqlite3.connect(self.database)
        c = conn.cursor()
        profileID = orbit
        with self.lock:
            c.execute('UPDATE ' + self.table + ' SET isProcessing=1 WHERE profileID=?',(profileID,) )
            conn.commit()
            conn.close()

    def set_entry_as_finished(self, orbit, failed=0):
        conn = sqlite3.connect(self.database)
        c = conn.cursor()
        profileID = orbit
        with self.lock:
            c.execute('UPDATE ' + self.table + ' SET isFinished=1 WHERE profileID=?',(profileID,) )
            c.execute('UPDATE ' + self.table + ' SET isProcessing=0 WHERE profileID=?',(profileID,) )
            c.execute('UPDATE ' + self.table + ' SET failed=? WHERE profileID=?',(failed,profileID) )
            conn.commit()
            conn.close()

    def reset_failed(self, new_priority=None, min_orbit=None, max_orbit=None):
        conn = sqlite3.connect(self.database)
        c = conn.cursor()
        data = c.execute('SELECT * FROM ' + self.table + ' WHERE isReady=1 AND isProcessing=0 AND failed=1 ORDER BY priority DESC')
        temp = data.fetchall()
        for entry in temp:
            orbit = int(entry[0])
            with self.lock:
                if (not min_orbit or orbit > min_orbit ) and (not max_orbit or orbit < max_orbit):
                    c.execute('UPDATE ' + self.table + ' SET isFinished=0 WHERE profileID=?',(orbit,) )
                    c.execute('UPDATE ' + self.table + ' SET isProcessing=0 WHERE profileID=?',(orbit,) )
                    c.execute('UPDATE ' + self.table + ' SET failed=0 WHERE profileID=?',(orbit,) )
                    if new_priority is not None:
                        c.execute('UPDATE ' + self.table + ' SET priority=? WHERE profileID=?',(new_priority,orbit) )
                    conn.commit()
        conn.close()

    def reset_isprocessing(self, new_priority=None, min_orbit=None, max_orbit=None):
        conn = sqlite3.connect(self.database)
        c = conn.cursor()
        data = c.execute('SELECT * FROM ' + self.table + ' WHERE isReady=1 AND isProcessing=1 AND failed=0 ORDER BY priority DESC')
        temp = data.fetchall()
        for entry in temp:
            orbit = int(entry[0])
            with self.lock:
                if (not min_orbit or orbit > min_orbit ) and (not max_orbit or orbit < max_orbit):
                    c.execute('UPDATE ' + self.table + ' SET isFinished=0 WHERE profileID=?',(orbit,) )
                    c.execute('UPDATE ' + self.table + ' SET isProcessing=0 WHERE profileID=?',(orbit,) )
                    if new_priority is not None:
                        c.execute('UPDATE ' + self.table + ' SET priority=? WHERE profileID=?',(new_priority,orbit) )
                    conn.commit()
        conn.close()

    def reset_orbit(self, orbit, new_priority):
        conn = sqlite3.connect(self.database)
        c = conn.cursor()
        with self.lock:
            c.execute('UPDATE ' + self.table + ' SET isFinished=0 WHERE profileID=?',(orbit,) )
            c.execute('UPDATE ' + self.table + ' SET isProcessing=0 WHERE profileID=?',(orbit,) )
            if new_priority is not None:
                c.execute('UPDATE ' + self.table + ' SET priority=? WHERE profileID=?',(new_priority,orbit) )
            conn.commit()
        conn.close()

    def view_table(self):
        conn = sqlite3.connect(self.database)
        c = conn.cursor()
        for row in c.execute('SELECT * FROM ' + self.table + ' ORDER BY priority DESC'):
            print(row)

        conn.close()

    def populate_from_folder(self, folder, orbitfun=lambda x: x):
        files = listdir(folder)
        for f in files:
            m = re.search('o([0-9]*)', f)
            if m:
                if len(m.group(1)) > 1:
                    orbit = int(m.group(1))
                    self.add_orbit_to_table(orbit, orbitfun(orbit))

