import os
import h5py
import julian
import numpy as np
import re
from datetime import datetime, timedelta
import xarray as xr
from scipy import interpolate
from dateutil.relativedelta import relativedelta
import warnings
from collections import OrderedDict


RSAS_CUTOFF_MJD = julian.to_jd(datetime(year=2013, day=25, month=4), fmt='mjd')
RSAS_BEFORE = 0.3
RSAS_AFTER = 0.2

FILTER_MJD_START = julian.to_jd(datetime(year=2013, day=25, month=11), fmt='mjd')
FILTER_MJD_END = julian.to_jd(datetime(year=2014, day=25, month=1), fmt='mjd')

VERSION = '0001'

def load_orbit(h5file, month_start_mjd, month_end_mjd):
    a_grid_index = range(0, 60)
    mjd = h5file['Geo/MJD'][:]

    if np.nanmean(mjd) > FILTER_MJD_START and np.nanmean(mjd) < FILTER_MJD_END:
        return None

    good = (mjd > month_start_mjd) & (mjd < month_end_mjd)
    mjd = mjd[good]

    ozone_num_den = h5file['DataUnbounded/OzoneNumDen'][:][np.ix_(a_grid_index, good)] / 6.022140857e17    # mol m-2
    ozone_num_den_error = h5file['DataUnbounded/OzoneNumDenError'][:][np.ix_(a_grid_index, good)] / 6.022140857e17 # mol m-2
    atmospheric_temperature = h5file['Data/BackgroundTemperature'][:][np.ix_(a_grid_index, good)]   # /K
    atmospheric_pressure = h5file['Data/BackgroundPressure'][:][np.ix_(a_grid_index, good)] / 100   # /hPa

    altitude = h5file['Geo/Altitude'][:][a_grid_index]                                         # [km]
    vertical_resolution = np.ones(np.shape(altitude)) * 1
    latitude = h5file['Geo/Latitude'][:][good]
    longitude = h5file['Geo/Longitude'][:][good]
    trop_alt = h5file['Data/TropopauseHeight'][:][good]

    # Do rsas correction
    mean_mjd = np.nanmean(mjd)
    if mean_mjd > RSAS_CUTOFF_MJD:
        rsas_alt = RSAS_AFTER
    else:
        rsas_alt = RSAS_BEFORE
    ozone_num_den = interpolate.interp1d(altitude, ozone_num_den, axis=0, bounds_error=False)(
        altitude + rsas_alt)

    ozone_num_den_error = interpolate.interp1d(altitude, ozone_num_den_error, axis=0, bounds_error=False)(
        altitude + rsas_alt)

    strat_col = np.ones(np.shape(mjd)) * np.nan
    for idx, t in enumerate(trop_alt):
        col_good = (atmospheric_pressure[:, idx] < 200) & (atmospheric_pressure[:, idx] > 1)
        strat_col[idx] = np.sum(ozone_num_den[col_good, idx]) * 1000
        ozone_num_den[altitude < t+1, idx] = np.nan

    ozone_num_den_error[np.isnan(ozone_num_den)] = np.nan

    good = np.full(np.shape(ozone_num_den), True, dtype=bool)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        mean_prof = ozone_num_den / np.nanmean(ozone_num_den, axis=1)[:, np.newaxis]
    outlier_check = np.diff(mean_prof, axis=0, n=2)

    lenx = np.size(outlier_check, axis=0)
    leny = np.size(outlier_check, axis=1)
    for idx in range(0, lenx):
        for idy in range(0, leny):
            if np.abs(outlier_check[idx, idy]) > 3:
                strat_col[range(max(0, idy-5), min(leny, idy+5))] = np.nan
                good[np.ix_(range(max(0, idx+1-5), min(lenx, idx+1+5)), range(max(0, idy-5), min(leny, idy+5)))] = False

    if np.any(strat_col*2244.15 > 500):
        # Still an outlier, forcibly remove it
        for out_idx in np.nonzero(strat_col*2244.15 > 500)[0]:
            idx_to_nan = range(max(0, out_idx-5), min(len(mjd), out_idx+5))
            strat_col[idx_to_nan] = np.nan
            good[:, idx_to_nan] = False
        print("----------------------Large Strat Col---------------------------------")

    # num_out = len(mjd) - np.sum(good)
    # if num_out > 0:
    #     print('Removed ' + str(num_out) + ' outliers')
    # if num_out > 30:
    #     print('Large Number of Out')

    ozone_num_den[~good] = np.nan
    ozone_num_den_error[~good] = np.nan

    orbit_data = xr.Dataset({'pressure': (['time', 'altitude'], atmospheric_pressure.T),
                             'latitude': (['time'], latitude),
                             'longitude': (['time'], longitude),
                             'ozone_concentration': (['time', 'altitude'], ozone_num_den.T),
                             'ozone_concentration_standard_error': (['time', 'altitude'], ozone_num_den_error.T),
                             'temperature': (['time', 'altitude'], atmospheric_temperature.T),
                             'vertical_resolution': (['altitude'], vertical_resolution),
                             'stratospheric_ozone_column': (['time'], strat_col)},
                            coords={'time': mjd-15020,
                                    'altitude': altitude})

    orbit_data['pressure'].attrs = {'standard_name': 'air_pressure',
                                        'description': 'pressure taken from MERRA and interpolated to the current geometrical altitude grid',
                                        'units': 'hPa'}

    orbit_data['latitude'].attrs = {'standard_name': 'latitude',
                                    'units': 'degree_north'}

    orbit_data['longitude'].attrs = {'standard_name': 'longitude',
                                     'units': 'degree_east'}

    orbit_data['ozone_concentration'].attrs = {'standard_name': 'mole_concentration_of_ozone_in_air',
                                                              'units': 'mol m-3',
                                                              'description': "mole concentration or number of moles per unit volume (molarity) of ozone",
                                                              'multiplication_factor_to_convert_to_molecules_percm3': '6.022140857E+17'}

    orbit_data['ozone_concentration_standard_error'].attrs = {'non_standard_name': 'standard_uncertainty_of_ozone_concentration',
                                                                             'long_name': 'standard_uncertainty_of_ozone_concentration',
                                                                             'multiplication_factor_to_convert_to_molecules_percm3': '6.022140857E+17',
                                                                             'units': 'mol m-3'}

    orbit_data['temperature'].attrs = {'standard_name': 'air_temperature',
                                           'description': 'temperature taken from MERRA and interpolated to the current geometrical altitude grid',
                                           'units': 'K'}

    orbit_data['vertical_resolution'].attrs = {'long_name': 'vertical_resolution',
                                               'non_standard_name': 'vertical_resolution',
                                               'description': 'averaged vertical resolution at each vertical grid point',
                                               'units': 'km'}

    orbit_data['time'].attrs = {'standard_name': 'time',
                                'calendar': 'standard',
                                'units': 'days since 1900-01-01 00:00:00'}

    orbit_data['altitude'].attrs = {'standard_name': 'altitude',
                                    'units': 'km'}

    orbit_data['stratospheric_ozone_column'].attrs = {'non_standard_name': 'stratosphere_mole_content_of_ozone',
                                                      'long_name': 'stratosphere_mole_content_of_ozone',
                                                      'description': 'stratospheric ozone column between 200 hPa and 1 hPa in mole concentration or number of moles per unit area (molarity) of ozone',
                                                      'units': 'mol m-2',
                                                      'multiplication_factor_to_convert_to_molecules_percm2': '6.022140857E+19',
                                                      'multiplication_factor_to_convert_to_DU': '2241.15'}

    return orbit_data

def make_mjd_map(root_folder):
    """
    Creates a dictionary used to map OMPS orbits to MJD's
    Parameters
    ----------
    l2_folder: string
        Folder of the OMPS orbital output files

    Returns
    -------
    mjd_orbit: dict
        Dictionary where mjd_orbit[orbit] = (min mjd of orbit, max mjd of orbit)
    """
    l2_folder = root_folder + '/L2/'
    mjd_orbit = dict()
    for f in os.listdir(l2_folder):
        m = re.search('([0-9]*).h5', str(f))
        if m is not None and len(m.group(1)) > 0:
            orbit = int(m.group(1))
            if orbit in mjd_orbit:
                continue
            else:
                try:
                    fi = h5py.File(l2_folder + f, "r")
                    mjd = fi['Geo/MJD'][:]
                    mjd_orbit[orbit] = (np.min(mjd), np.max(mjd))
                except OSError:
                    pass
    return mjd_orbit


def monthly_file_name(mjd):
    day_start = julian.from_jd(mjd, fmt='mjd')

    return 'ESACCI-OZONE-L2-LP-OMPS_LP_SUOMI_NPP-SASK_2D_V1_0_2_HARMOZ_ALT-' + day_start.strftime("%Y%m") + '-fv' + VERSION + '.nc'


if __name__ == "__main__":
    root_folder = r'Z:/data/OMPS/OMPS_Tomography_v1.0.2/'

    mjd_orbit = make_mjd_map(root_folder)

    mjd_start = int(np.floor(np.nanmin([m[0] for o, m in mjd_orbit.items()])))
    mjd_end = int(np.ceil(np.nanmax([m[1] for o, m in mjd_orbit.items()])))

    date_start = julian.from_jd(mjd_start, fmt='mjd')
    date_start = datetime(year=date_start.year, month=date_start.month, day=1)

    date_end = julian.from_jd(mjd_end, fmt='mjd')
    date_end = datetime(year=date_end.year, month=date_end.month+1, day=1)

    one_month = relativedelta(months=+1)

    mjd = mjd_start
    current_month = date_start
    while mjd < mjd_end:
        end_month = current_month + one_month
        month_end_mjd = julian.to_jd(end_month, fmt='mjd')
        # Process the data
        orbits = [o for o, m in mjd_orbit.items() if m[1] > mjd and m[0] < month_end_mjd]

        if len(orbits) > 0:
            monthly_data = None
            for orbit in sorted(orbits):
                try:
                    orbit_file = h5py.File(root_folder + '/L2/o' + str(orbit) + '.h5', 'r')
                    raw_orbit_file = h5py.File(root_folder + '/raw/Retinfo_' + str(orbit) + '.h5', 'r')
                    chi_squared = raw_orbit_file['RetInfo/Ozone_chi_squared'].value
                    if chi_squared < 2:
                        orbital_data = load_orbit(orbit_file, mjd, month_end_mjd)
                        if orbital_data is None:
                            continue
                        if monthly_data is None:
                            monthly_data = orbital_data
                        else:
                            monthly_data = xr.concat((monthly_data, orbital_data), dim='time')
                except KeyError:
                    pass

            def str_t_func(dt):
                return dt.strftime("%Y%m%dT%H%M%SZ")
            str_format = 'YYYYMMDDThhmmssZ'

            creation_date = datetime.utcnow()

            name = monthly_file_name(mjd)
            if monthly_data is not None:
                monthly_med = monthly_data['ozone_concentration'].median(dim='time')
                MAD = (np.abs(monthly_data['ozone_concentration'] - monthly_med)).median(dim='time')

                mod_z = 0.6745 * (monthly_data['ozone_concentration'] - monthly_med) / MAD

                num_before = np.sum(~np.isnan(monthly_data['ozone_concentration'].values.flatten()))

                monthly_data['ozone_concentration'].values[mod_z.values > 4] = np.nan
                monthly_data['ozone_concentration_standard_error'].values[mod_z.values > 4] = np.nan

                mod_z_col = np.nanmax(mod_z, axis=1)
                monthly_data['stratospheric_ozone_column'].values[mod_z_col > 4] = np.nan

                num_after = np.sum(~np.isnan(monthly_data['ozone_concentration'].values.flatten()))

                print(str(num_before - num_after) + ' Outliers removed')

                monthly_data.attrs = OrderedDict((('title', 'ESA CCI ozone product level 2'),
                                                  ('institution', 'University of Saskatchewan'),
                                                  ('history', 'Product generated with SASKTRAN'),
                                                  ('platform', 'Suomi-NPP'),
                                                  ('sensor', 'OMPS-LP'),
                                                  ('affiliation', 'University of Saskatchewan'),
                                                  ('date_created', str_t_func(creation_date)),
                                                  ('creator_name', 'Daniel Zawada'),
                                                  ('creator_url', 'www.odin-osiris.usask.ca'),
                                                  ('creator_email', 'daniel.zawada@usask.ca'),
                                                  ('address', '116 Science Place, Saskatoon, SK, Canada, S7N 5E2'),
                                                  ('naming_authority', 'University of Saskatchewan'),
                                                  ('file_version', 'fv' + VERSION),
                                                  ('file_version_description', 'fv0001 Initial Version'),
                                                  ('level_1_data_version', 'V2.0-V2.4'),
                                                  ('level_2_data_version', 'V1.0.2'),
                                                  ('file_id', name),
                                                  ('comment', 'These data were produced at the University of Saskatchewan as part of the ESA OZONE CCI project'),
                                                  ('summary', 'Level 2 OMPS-LP ozone profiles'),
                                                  ('keywords', 'ozone,satellite,remote sensing,observation,atmosphere,omps,cci'),
                                                  ('value_for_nodata', 'NaN'),
                                                  ('Conventions', 'CF-1.6'),
                                                  ('standard_name_vocabulary', 'NetCDF Climate and Forecast(CF) Metadata Convention version 18'),
                                                  ('license', 'ESA CCI Guidelines'),
                                                  ('restriction', 'Restricted under the use of ozone cci guidelines'),
                                                  ('geospatial_lat_min', "-82.0"),
                                                  ('geospatial_lat_max', "82.0"),
                                                  ('geospatial_lon_min', "0"),
                                                  ('geospatial_lon_max', "360"),
                                                  ('geospatial_vertical_min', '0.5 km'),
                                                  ('geospatial_vertical_max', '59.5 km'),
                                                  ('string_date_format', str_format),
                                                  ('time_coverage_start', str_t_func(date_start)),
                                                  ('time_coverage_end', str_t_func(end_month - timedelta(seconds=1))),
                                                  ('number_of_altitude_levels', "60"),
                                                  ('number_of_profiles', str(len(monthly_data['time'].values))),
                                                  ('tracking_id', "d3850ffa-60c8-442a-adbb-28bb944445d3")))
                # monthly_data.to_netcdf(root_folder + '/HARMOZ/' + name, format='NETCDF4_CLASSIC')
                monthly_data.to_netcdf(r'\\utls\\OsirisDataProducts\\Level2\\CCI\\' + name, format='NETCDF4_CLASSIC')

        # Move to next month
        current_month = end_month
        mjd = julian.to_jd(current_month, fmt='mjd')