import numpy as np
from scipy import sparse as sparse
from radtran import options as skopt, geometry
from radtran.sasktran import SasktranGeneric
from sasktranif import sasktranif as skif
from tomography.species import grid as grid
import gc


class SasktranHRTwoDim(SasktranGeneric):
    """
    A specialized version of SASKTRAN HR which supports calculations with
    a two dimensional atmosphere.  This class is designed for use in tomographic
    retrievals, for general two/three dimensional calculations it is advised to
    use the SasktranHR class.
    """
    def __init__(self, atmosphere, options=None, maxdifference_seconds=100):
        self.wavel_nm = None
        self.options = options
        self.atmosphere = atmosphere
        self.numlos = 0
        self.is_engine_initialized = False

        if options is None:
            # Set up some default options
            self.options = skopt.HROptions()

            self.options.add_option('numordersofscatter', 10)
            self.options.add_option('raytracingshells', 1000)
            self.options.add_option('numdiffuseprofilesinplane', 5)
            self.options.add_option('solarraytracingshells', 1000)
            self.options.add_option('integrationtechnique', 0)
            self.options.add_option('wfprecision', 1)
        else:
            self.options = options

        # These options are required for the class to work
        self.options.add_option('diffuseplacementtype', 0)
        self.options.add_option('calcwf', 3)
        self.options.add_option('opticaltabletype', 2)

        self.maxmjdidfference = maxdifference_seconds / (60 * 60 * 24)  # Seconds
        self.los = []
        self.obs = []
        self.mjd = []
        self.wfsparsity = None
        self.splits = []
        self.retrieval_grid = None
        self.scanslices = None
        self.numwfalt = 100

    def add_line_of_sight(self, observer, look, mjd):
        self.los.append(look)
        self.obs.append(observer)
        self.mjd.append(mjd)
        self.numlos += 1

    def setup_geometry(self, scanslices, placementtype='scan', numinbetween=1):
        # Calculate the tangent points
        tp, tanalts = geometry.tangent_points(self.obs, self.los)

        # Using the tangent altitudes, calculate the scan slices
        self.scanslices = scanslices
        self.mjd = np.array(self.mjd)
        self.los = np.array(self.los)
        self.setup_splits()

        # TODO: This might be moved into the 2D retrieval species
        self.retrieval_grid = grid.TwoDimGrid(tp, self.los, self.mjd, self.scanslices, placementtype=placementtype,
                                              numinbetween=numinbetween)

    def setup_splits(self):
        """
        Sets up the radiative transfer calculation splits.  Since the radiative
        transfer model assumes one sun position, we cannot add all of the
        lines of sight to the model at the same time.  The parameter maxmjddifference
        is used to combine scans together that do not vary by a user specified
        mjd difference
        """
        self.splits = []
        currentidx = 0

        if len(self.scanslices) > 1:
            currentmjd = np.mean(self.mjd[self.scanslices[0]])

            for idx in range(1, len(self.scanslices)):
                if (np.mean(self.mjd[self.scanslices[idx]]) - currentmjd) > self.maxmjdidfference:
                    self.splits.append(range(currentidx, idx))
                    currentidx = idx
                    currentmjd = np.mean(self.mjd[self.scanslices[idx]])
            self.splits.append(range(currentidx, len(self.scanslices)))

    def set_wf_alts(self, alts, widths):
        self.options.add_option('wfheights', alts)
        self.options.add_option('wfwidths', widths)

        self.numwfalt = len(alts)

    def calculate_radiance(self):
        """
        Calculates the radiance note that the return format for the
        weighting functions is slightly different than SasktranHR for
        exmaple

        Returns
        -------
        rad
            2d numpy array, (los, wavelength)
        wf
            List of length (wavelength) of sparse matrices of shape (los, perturbation)
        """
        numlos = np.size(self.mjd)
        numwavel = np.size(self.wavel_nm)
        numpert = self.retrieval_grid.numretprof * self.numwfalt
        rad = np.zeros((numwavel, numlos))

        # Store the weighting functions using a sparse matrix, linked list format
        # is good for the construction
        wf = [None] * numwavel
        for idwav in range(numwavel):
            wf[idwav] = sparse.lil_matrix((numlos, numpert))
        # TODO: The sparsity pattern will be roughly the same for every wavelength,
        # so we could only make the first one for now

        for idx in range(len(self.splits)):
            # logging.info(self.splits[idx])
            meantp = np.mean(self.retrieval_grid.averagescantp[self.splits[idx], :], axis=0)
            engine = skif.ISKEngine('HR')
            # We have to give the meantp to the atmosphere unfortunately
            self.atmosphere.add_to_engine(engine, meantp)

            self.options.add_to_engine(engine)
            engine.SetWavelengths(self.wavel_nm)
            numlos = 0
            for idy in range(np.size(self.splits[idx])):
                for s in self.scanslices[self.splits[idx][idy]]:
                    engine.AddLineOfSight(self.mjd[s][0], self.obs[s], self.los[s])
                    numlos += 1
            engine.InitializeModel()
            angleidx = self.atmosphere.angleidx

            losidx_raw = np.array(self.scanslices)[self.splits[idx]]
            losidx = [item for sublist in losidx_raw for item in sublist]
            losidx = np.array(losidx)

            for idwav in range(numwavel):
                engine.SetWavelengths([self.wavel_nm[idwav]])
                ok, radt = engine.CalculateRadiance()
                rad[idwav, losidx] = np.squeeze(radt)

                if self.options.optmap['calcwf'] != 0:
                    wfidxt = slice(angleidx[0] * self.numwfalt, (angleidx[-1] + 1) * self.numwfalt)
                    for k in range(numlos):
                        ok, t = engine.GetProperty('wf[' + str(k) + ']')
                        wf[idwav][losidx[k], wfidxt] = t

            # We force a garbage collection here since on machines like westgrid where the OS memory limit is not
            # the same as the actual provided limit the garbage collector might be too passive
            gc.collect()

        # Now convert to csr format for faster matrix operations
        for idwav in range(numwavel):
            wf[idwav] = sparse.csr_matrix(wf[idwav])

        return np.transpose(rad), wf
