import copy as copy

import numpy as np
from scipy import signal

import retrieval.species as species
import utility.solar_spectrum as solar
import radtran.geometry as geometry


class TomographicAlbedo(species.Species):
    """
    Class to perform the tomographic albedo calculation. We take high altitude
    measurements and assume that the radiance at this altitude is primarily influenced
    by the ground directly underneath it.
    """

    def __init__(self, atmosphere, smooth_factor=1):
        self.atmosphere = atmosphere
        self.desired_wavel = np.array([745.67])
        self.alt_range_idx = 5  # The width of altitudes to use in the measurement vector
        self.highest_alt_m = 42000  # The highest altitude to use in the measurement vector

        # In configure_instrument_and_method this is constructed with
        # a tuple (low_alt, high_alt) based on the available measurements
        self.alt_range = []

        self.y0 = None
        self.scanalbedo = []
        self.smoothfactor = smooth_factor
        self.ident = 'tomoalbedo'
        self.grid = None

        self.tp = None
        self.retlat = None
        self.retlon = None
        self.scanlat = None
        self.scanlon = None
        self.tanalt = None
        self.ret_wavel = None
        self.spectrum = None

    def get_required_wavelengths(self, inst):
        """
        Calculates the required wavelengths for the retrieval

        Parameters
        ----------
        inst:
            Instrument, should have wavel_nm parameter

        Returns
        -------
        wavel : numpy array
            Wavelengths required for the retrieval
        """
        wavel_nm = []
        for w in self.desired_wavel:
            wave_idx = np.argmin(np.abs(np.unique(inst.wavel_nm) - w))
            wavel_nm.append(inst.wavel_nm[wave_idx])
        wavel_nm = np.unique(np.array(wavel_nm))
        return wavel_nm

    def measurement_vector(self, inst, doerror=False):
        """
        Measurement vector is the sum of radiances at wavelengths in
        self.desired_wavel averaged over the ranges specified in self.alt_range
        """
        rad = inst.rad

        scanslices = inst.radtran.scanslices

        y = np.zeros(len(scanslices))
        for idx in range(len(scanslices)):
            s = scanslices[idx]
            # Find the normalization altitudes
            for idy, w in enumerate(self.ret_wavel):
                lowidx = (np.asarray(self.tanalt)[s] < self.alt_range[idx][0]).nonzero()[0][-1]
                highidx = (np.asarray(self.tanalt)[s] < self.alt_range[idx][1]).nonzero()[0][-1]
                widx = np.argmin(np.abs(inst.wavel_nm - w))
                radR = np.mean(rad[s[lowidx:highidx], widx])
                if radR > 1:
                    radR /= self.spectrum[idy]
                y[idx] += radR
        if self.y0 is None:
            return y
        else:
            return y - self.y0, None

    def measurement_vector_jacobian(self, inst):
        pass

    def interp(self):
        """
        Takes the albedo from the measurement grid and converts it to the retrieval grid
        """
        return self.grid.interpolate_onto_grid(self.scanlon, self.scanlat, [], self.scanalbedo, False)

    def interp_back(self):
        """
        Takes the albedo on the retrieval grid and converts it back to the measurement grid
        """
        return self.grid.interpolate_from_grid(self.scanlon, self.scanlat, [], self.atmosphere.twodimalbedo, False)

    def configure_instrument_and_method(self, ret_inst, meas_inst, method):
        """
        Sets up the species for use in the retrieval.  y0 is calculated by setting
        the albedo to 0
        """
        # determine the highest altitude that we can use for the albedo retrieval
        # In the early years, nan's were reported at higher altitudes which can cause
        # problems if we just use the 37 - 42 km range for example
        self.grid = meas_inst.radtran.retrieval_grid
        widx = np.argmin(np.abs(meas_inst.wavel_nm - self.desired_wavel[0]))
        inst_rad = meas_inst.rad[:, widx]

        self.tp, self.tanalt = geometry.tangent_points(ret_inst.observer, ret_inst.los)

        self.tanalt = np.array(self.tanalt)
        self.scanlat = np.zeros(len(ret_inst.radtran.scanslices))
        self.scanlon = np.zeros(len(ret_inst.radtran.scanslices))
        # Calculate the latitudes, longitudes of the scans in the retrieval range
        for idx, s in enumerate(ret_inst.radtran.scanslices):
            scan_rad = inst_rad[s]
            highidx = np.nonzero((self.tanalt[s] < self.highest_alt_m) & (~np.isnan(scan_rad)))[0][-1]
            lowidx = highidx - self.alt_range_idx

            self.alt_range.append((self.tanalt[s][lowidx], self.tanalt[s][highidx]))

            tpslice = slice(s[0], s[-1])

            mean_tp = np.mean(np.array(self.tp[tpslice]), axis=0)

            self.scanlat[idx], self.scanlon[idx], alt = geometry.lat_lon_alts([mean_tp])

        self.retlat, self.retlon, alt = geometry.lat_lon_alts(meas_inst.radtran.retrieval_grid.retprofileloc)
        self.retlon[self.retlon > 180] -= 360
        self.ret_wavel = self.get_required_wavelengths(meas_inst)

        ret_inst.set_wavelengths(self.ret_wavel)

        self.spectrum = solar.convolved_solar_spectrum(self.ret_wavel, meas_inst.wavel_nm, meas_inst.psf)

        # Calculate y0 using ret_inst
        oldprof = copy.copy(self.atmosphere.twodimalbedo)
        self.atmosphere.twodimalbedo = np.zeros(np.shape(self.atmosphere.twodimalbedo))
        ret_inst.calculate_radiance()
        self.y0 = None
        self.y0 = self.measurement_vector(ret_inst)
        self.atmosphere.twodimalbedo = oldprof

        if method.needs_jacobian():
            raise Exception('Error: Trying to use a method which requires jacobian in TomographicAlbedo')

        self.scanalbedo = self.interp_back()

    def get_x(self):
        return self.scanalbedo

    def set_x(self, profile):
        self.scanalbedo = profile

    def update_profile(self, delta, mode='multiplicative'):
        self.scanalbedo *= delta
        t = self.interp()
        if self.smoothfactor == 1:
            self.atmosphere.twodimalbedo = t
        else:
            self.atmosphere.twodimalbedo = signal.resample(t, len(t),
                                                           window=('general_gaussian', 1, len(t) / self.smoothfactor))
        self.atmosphere.twodimalbedo[self.atmosphere.twodimalbedo > 1] = 1
        self.atmosphere.twodimalbedo[self.atmosphere.twodimalbedo < 0] = 0

    def output_data(self):
        return self.atmosphere.twodimalbedo
